 /*
		misc DSP functions
		
    Copyright (C) 2015  Ralph Schenn

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h> 
#include "Debug.h" 
#include "Matrix.h" 
#include "ROM.h" 
#include "Stack.h" 

/* downsample array by factor 
input signal 
length of input signal
downsampling factor 
output signal 
length of output is floor (l / f)
even simple this is much faster than basic (especially for long lists)
doing this on the stack directly isnt much faster (because lst has to reversed)
http://www.eetimes.com/document.asp?doc_id=1275556&page_number=1 */
void Downsamp(double x[],
	unsigned short l,
	unsigned short f,
	double y[])
{
	unsigned short i, j = 0, k = 1;
	
	// loop over input samples
	for (i = 0; i < l; i++) {
		// copy only every fth sample from input to output
		if (k == 1) {
			y[j++] = x[i];
		}
		
		// increment or reset k
		k = k < f ? k + 1 : 1;
	}	
}

/* samples per period */
double SampPerPeriod(double fs, 
	double f)
{
	// for Nyquist frequency the period length is PI
	// (corresonding to 2 samples)
	return 1 / (f / fs);
}

#define COLS 3
#define FREQ 0
#define AMPL 1
#define ANGL 2

/* generate discrete superposition of sine waves 
sampling frequency 
number of samples 
[freq, amp, angle;...] (one row per frequency) 
??? will be faster when calculating only one period for lowest freq 
and duplicating the samples for following periods (n may not be integer 
multiple ob base frq period length */
void supossin(void)
{
	Access_AMS_Global_Variables;

	EStackIndex argptr = top_estack;
	double fs, fn, f, a, p, fmin = FLOATTAB[FPI_POSINF], fmax = 0;
	volatile HANDLE faaHdl, sHdl;
	double *faa, *sPtr;
	unsigned short n, r, c, i, j;
	char msg[MAX_HELP_STR];

	CheckNumArgs (3);

	// get sampling frequency
	fs = ForceFloat (argptr);

	if (fs <= 0) {
		ER_throwVar (ER_DOMAIN);
	}
	
	// Nyquist frequency
	fn = fs / 2;

	argptr = next_expression_index (argptr);

	// get number of samples
	n = (unsigned short)GetValue (argptr, 1, MAX_LST_LEN);

	argptr = next_expression_index (argptr);

	TRY
		// get matrix with freq, ampl, angle
		GetFloatMatArg (argptr, &faaHdl, &r, &c);
		
		if (c != 3) {
			ER_throwVar (ER_DOMAIN);
		}
	
		// allocate array with output signal
		sHdl = HeapAllocThrow (n * sizeof (double));

		faa = HeapDeref (faaHdl); 
		sPtr = HeapDeref (sHdl); 

		// initialize output signal
		ArrInit (sPtr, n);

		// loop over rows of matrix with sinusoids (freq, ampl, phase)
		for (i = 0; i < r; i++) {
			f = faa[IDX2D (COLS, i, FREQ)];
			
			// no negative frequencies
			// no frequencies higher than Nyquist frequnency
			if (f < 0 || f > fn) {
				ER_throwVar (ER_DOMAIN);
			}
			
			// actual frequency is lower
			if (f < fmin) {
				fmin = f;
			}
			
			// actual frequency is higher
			if (f > fmax) {
				fmax = f;
			}
			
			a = faa[IDX2D (COLS, i, AMPL)];
			p = faa[IDX2D (COLS, i, ANGL)];
			
			// loop over output samples
			for (j = 0; j < n; j++) {
				// calc value for current frequency and add to sum of previous freqs
				sPtr[j] += a * sin (FLOATTAB[FPI_TWOPI] * f * j / fs + p);
			}
		}
        
       	// displaysampled periods for min and max frequency
       	sprintf (msg, 
       		"PER FMIN %.2f, FMAX %.2f", 
       		n / SampPerPeriod (fs, fmin), 
       		n / SampPerPeriod (fs, fmax));
       		
       	ST_helpMsg (msg);

		// return superposition of all freqs in matrix
		push_float_list (sPtr, n);
	FINALLY
		HeapFreeIndir ((HANDLE *)&faaHdl); 
		HeapFreeIndir ((HANDLE *)&sHdl); 
	ENDFINAL
}

/* downsample array by factor 
input signal 
length of input signal
downsampling factor 
output signal 
length of output is floor (l / f)
even simple this is much faster than basic (especially for long lists)
doing this on the stack directly isnt much faster (because lst has to reversed)
http://www.eetimes.com/document.asp?doc_id=1275556&page_number=1 */
void Upsamp(double x[],
	unsigned short l,
	unsigned short f,
	double y[])
{
	unsigned short i, j = 0, k;
	
	// loop over input samples
	for (i = 0; i < l; i++) {
		y[j++] = x[i];
		
		// loop over zeros that have to be inserted
		for (k = 1; k < f; k++) {
			y[j++] = 0;
		}
	}	
}
