//   fft.h - declaration of class
//   of fast Fourier transform - FFT
//
//   The code is property of LIBROW
//   You can use it on your own
//   When utilizing credit LIBROW site
//   
//   The following modifications (c) 2015 Ralph Schenn
//   - port from C++ to C
//   - port to TI-89 pocket calculator
 
#include <tiams.h>
#include "Complex.h"
#include "Debug.h"
#include "FFT.h"
#include "Matrix.h"
#include "Real.h"
#include "ROM.h"
#include "Stack.h"
 
#define NUM_ARGS_MIN 2
#define NUM_ARGS_MAX 3
#define DIR 1
#define LEN 2

/* fast fourier transform */
void fft(void)
{	
	Access_AMS_Global_Variables; 

	EStackIndex argptr = top_estack;
    int dir;
	volatile HANDLE handle;
	Complex *array;
    unsigned short larg = 0, lold, llst;
    char help[MAX_HELP_STR]; 
    unsigned char numArgs;
    
	// check number of parameters
	numArgs = CheckNumArgsInt (NUM_ARGS_MIN, NUM_ARGS_MAX);
		
	// get second argument
	dir = GetSignShortArgPos (DIR);
	
	// check value of second argument
	if (abs (dir) != 1) {
		ER_throwVar (ER_DOMAIN);
	}
	
	// length was passed as third argument
	if (numArgs == 3) {
	    larg = GetSignShortArgPos (LEN);
	}			
		
	TRY	
		// read list into array
		llst = GetComplexFloatListArg (argptr, &handle);
				
		// list is empty
		if (llst == 0) {
			ER_throwVar (ER_DIMENSION);
		}
		
		// output cant be shorter than input
		if (numArgs == 3 && larg < llst) {
			ER_throwVar (ER_DOMAIN);
		}	
		
		lold = llst;
		
		// pad to next smallest power of two if nevessary
		llst = CPadPow2 (handle, llst, larg);
		
		// original length was not a power of two
		if (llst > lold) {
			sprintf (help, "OLD LENGTH %d NEW LENGTH %d", lold, llst);
			ST_helpMsg (help); 
		}
		
		array = HeapDeref (handle);
	
  		//CArrPause (array, l);
		
		// perform fourier transform
  		FFT (array, llst, dir == -1 ? TRUE : FALSE);

  		//CArrPause (array, l);

		// push list in reverse order to estack
		push_complex_float_list (array, llst); 			
	FINALLY
		// avoid address error when arg is empty list
		HeapFreeIndir ((HANDLE *)&handle);
	ENDFINAL
}

/* Inplace version of rearrange function */
static void Rearrange(Complex Data[], 
	const unsigned short N)
{
	unsigned short Target = 0, Position, Mask;

	// Swap position
 	Complex Temp; 				

	// Process all positions of input signal
	for (Position = 0; Position < N; ++Position)
	{
		//   Only for not yet swapped entries
		if (Target > Position)
		{
			//   Swap entries
			Temp = Data[Target];
			Data[Target] = Data[Position];
			Data[Position] = Temp;
		}
 
		// Bit mask
		Mask = N;
 
		// While bit is set
        while (Target & (Mask >>= 1))
		{
			// Drop bit
			Target &= ~Mask;
		}
 
		// The current bit is 0 - set it
		Target |= Mask;
	}
}

/* FFT implementation */
static void Perform(Complex Data[], 
	const unsigned short N, 
	const Boolean Inverse)
{
	Access_AMS_Global_Variables; 
	
	unsigned short Step, Jump, Group, Pair, Match;	
	const double pi = Inverse ? FLOATTAB[FPI_ONEPI] : -FLOATTAB[FPI_ONEPI]; 
	double delta, Sine;	
	Complex Multiplier, Factor, Product;

	// Iteration through dyads, quadruples, octads and so on...
	for (Step = 1; Step < N; Step <<= 1)
	{
		// Jump to the next entry of the same transform factor
		Jump = Step << 1;
 
		// Angle increment
		delta = pi / (double)Step;
 
		// Auxiliary sin(delta / 2)
		Sine = sin (delta * .5);
 
		// Multiplier for trigonometric recurrence
		Multiplier.real = -2. * Sine * Sine;
		Multiplier.imag = sin (delta);
 
		// Start value for transform factor, fi = 0
		Factor = COne;
		 								
		// Iteration through groups of different transform factor
		for (Group = 0; Group < Step; ++Group)
		{
			// Iteration within group
			for (Pair = Group; Pair < N; Pair += Jump)
			{
				// Match position
				Match = Pair + Step;
 
				// Second term of two-point transform
				Product = CMul (Factor, Data[Match]);
 
				// Transform for fi + pi
				Data[Match] = CSub (Data[Pair], Product);
 
				// Transform for fi
				Data[Pair] = CAdd (Data[Pair], Product);
			}
 
			// Successive transform factor via trigonometric recurrence
			Factor = CAdd (CMul (Multiplier, Factor), Factor);
		}
	}
}

/* Scaling of inverse FFT result */
static void ScaleData(Complex Data[], 
	const unsigned short N)
{
	unsigned short Position;
	Complex Factor;

	Factor.real = 1. / (double)N;
	Factor.imag = 0.;        
 
	// Scale all data entries
	for (Position = 0; Position < N; ++Position)
	{
		Data[Position] = CMul (Data[Position], Factor);
	}
}

/* FOURIER TRANSFORM, INPLACE VERSION
Data  - both input data and output
N     - length of both input data and result */
Boolean FFT(Complex Data[],
	const unsigned short N,
	const Boolean Inverse)
{
	// Check input parameters
	if (!Data || N < 1 || N & (N - 1))
	{
		return FALSE;
	}
 
	// Rearrange
	Rearrange(Data, N);
 
	// Call FFT implementation
	Perform(Data, N, Inverse);
 
	// Scale if necessary
	if (Inverse)
	{
		ScaleData(Data, N);
	}
 
	// Succeeded
	return TRUE;
}
  
/* non-destructive fourier transform */
HANDLE FFTNonDes(Complex Data[],
	const unsigned short N,
	const Boolean Inverse)
{
	Complex *rPtr;
	HANDLE rHdl = HeapAllocThrow (N * sizeof (Complex)); 

	rPtr = HeapDeref (rHdl);
		
	memcpy (rPtr, Data, N * sizeof (Complex)); 
	
	FFT (rPtr, N, Inverse);
	
	return rHdl;
}
